%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Module random_mt
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% @copyright 2007 Tim Bates
%
% @author Tim Bates <tim@bates.id.au>
%
% @doc  The Mersenne Twister PRNG in Erlang.
%
%       The Mersenne Twister is a fast, high-quality pseudo-random number
%       generator (PRNG) with a colossal period of 2^19937-1, a very high order
%       of dimensional equidistribution (ie negligible serial correlation
%       between successive output values) and excellent statistical randomness
%       properties.
%
%       This is a pure Erlang port of the C version published at
%       [http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html]
%
% @version 1.0 (2007-06-06)
%

-module(random_mt).

-export([
    % mt19337ar.c interface
    init/1,
    genrand/1,
    genrand_int32/1,
    genrand_int31/1,
    genrand_real1/1,
    genrand_real2/1,
    genrand_real3/1,
    genrand_res53/1,
    % stdlib random interface
    seed/0,
    seed/3,
    seed0/0,
    uniform/0,
    uniform/1,
    uniform_s/1,
    uniform_s/2
]).



% Period parameters
-define(N, 624).
-define(M, 397).
-define(MATRIX_A, 16#9908b0df).   % constant vector a
-define(UPPER_MASK, 16#80000000). % most significant w-r bits
-define(LOWER_MASK, 16#7fffffff). % least significant r bits

% @type ran_mt(). Random state used to generate the next random number.


%
% @spec init(Seed) -> ran_mt()
%       Seed = integer() | list() | tuple()
%
% @doc  Seeds the random number generator, returning the initial state.
%       Seed must be a 32-bit integer or a list or tuple of same.
%
init(Seed) when is_list(Seed) ->
    init_by_array(Seed);
init(Seed) when is_tuple(Seed) ->
    init_by_array(tuple_to_list(Seed));
init(Seed) when Seed >= 0, Seed =< 16#FFFFFFFF ->
    InitialList = populate_initial_array(1, [Seed]),
    {[], InitialList}.



% initializes mt[N] with a seed
populate_initial_array(?N, List) ->
    lists:reverse(List);

populate_initial_array(Mti, List) ->
    NewVal = (1812433253 * (hd(List) bxor (hd(List) bsr 30)) + Mti) band 16#FFFFFFFF, 
    populate_initial_array(Mti + 1, [NewVal | List]).


init_by_array(InitKey) when is_list(InitKey) ->
    {[], List} = init(19650218),
    %k = (N>key_length ? N : key_length);
    K = if ?N > length(InitKey) -> ?N; true -> length(InitKey) end,

    %i=1; j=0;
    %for (; k; k--)
    {I, List1, InitList1, Acc1} = mix_key(K, 1, 0, InitKey, InitKey, tl(List), List, [hd(List)]),

    %for (k=N-1; k; k--)
    List2 = mix_state(?N-1, I, List1, InitList1, Acc1),

    %mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */ 
    List3 = [ 16#80000000 | tl(List2) ],

    {[], List3}.



%for (; k; k--)
mix_key(0, I, _, _, _, List, InitList, Acc) ->
    {I, List, InitList, Acc};
mix_key(K, I, J, InitKey, InitKeyReset, List, InitList, Acc) ->
    MtI = hd(List),
    MtIt1 = hd(Acc),

    %mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1664525UL))
    %        + init_key[j] + j; /* non linear */
    %mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
    NewMtI = ((MtI bxor ((MtIt1 bxor (MtIt1 bsr 30)) * 1664525)) + hd(InitKey) + J) band 16#FFFFFFFF,
    Acc2 = [ NewMtI | Acc ],

    %i++; j++;
    %if (i>=N) { mt[0] = mt[N-1]; i=1; }
    %if (j>=key_length) j=0;
    {NewList, NewInitList, NewAcc, NewI} = if
        I + 1 >= ?N ->
            NewIL = [ NewMtI | tl(lists:reverse(Acc2)) ],
            {tl(NewIL), NewIL, [NewMtI], 1};
        true ->
            {tl(List), InitList, Acc2, I+1}
    end,
    case tl(InitKey) of
        [] ->
            mix_key(K - 1, NewI, 0, InitKeyReset, InitKeyReset, NewList, NewInitList, NewAcc);
        NewInitKey ->
            mix_key(K - 1, NewI, J+1, NewInitKey, InitKeyReset, NewList, NewInitList, NewAcc)
    end.

%for (k=N-1; k; k--)
mix_state(0, _, List, _, []) ->
    List;
mix_state(0, _, List, _, [Elem | Acc]) ->
    mix_state(0, x, [Elem | List], x, Acc);
mix_state(K, I, List, InitList, Acc) ->
    MtI = hd(List),
    MtIt1 = hd(Acc),

    %mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1566083941UL))
    %       - i; /* non linear */
    %mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
    NewMtI = ((MtI bxor ((MtIt1 bxor (MtIt1 bsr 30)) * 1566083941)) - I) band 16#FFFFFFFF,
    Acc2 = [ NewMtI | Acc ],

    %i++;
    %if (i>=N) { mt[0] = mt[N-1]; i=1; }
    {NewList, NewInitList, NewAcc, NewI} = if
        I + 1 >= ?N ->
            NewIL = [ NewMtI | tl(lists:reverse(Acc2)) ],
            {tl(NewIL), NewIL, [NewMtI], 1};
        true ->
            {tl(List), InitList, Acc2, I+1}
    end,
    mix_state(K - 1, NewI, NewList, NewInitList, NewAcc).



%
% @spec genrand(State::State) -> {Int32, NewState}
%       State = ran_mt()
%       Int32 = integer()
%       NewState = ran_mt()
%
% @doc  Generates a random number on [0,0xffffffff]-interval
%
genrand({[], InitList}) ->
    %for (kk=0;kk<N-M;kk++)
    {List1, Acc1} = mix1(0, InitList, lists:nthtail(?M, InitList), []),

    %for (;kk<N-1;kk++)
    List2 = mix2(?N-?M, List1, lists:reverse(Acc1), Acc1),

    Mt0 = hd(List2),
    MtMt1 = lists:nth(?M, List2),
    MtNt1 = lists:nth(?N, List2),
    %y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
    Y = (MtNt1 band ?UPPER_MASK) bor (Mt0 band ?LOWER_MASK),

    %mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];
    NewMtNt1 = MtMt1 bxor (Y bsr 1) bxor mag01(Y band 1),
    List3 = lists:reverse([NewMtNt1 | tl(lists:reverse(List2))]),

    genrand({List3, List3});

genrand({List, InitList}) ->
    %y = mt[mti++];
    Y = hd(List),

    % Tempering */
    %y ^= (y >> 11);
    Y1 = Y bxor (Y bsr 11),
    %y ^= (y << 7) & 0x9d2c5680UL;
    Y2 = Y1 bxor ((Y1 bsl 7) band 16#9d2c5680),
    %y ^= (y << 15) & 0xefc60000UL;
    Y3 = Y2 bxor ((Y2 bsl 15) band 16#efc60000),
    %y ^= (y >> 18);
    Y4 = Y3 bxor (Y3 bsr 18),

    {Y4, {tl(List), InitList}}.


% mag01[x] = x * MATRIX_A  for x=0,1
mag01(0) -> 0;
mag01(1) -> ?MATRIX_A.

%for (kk=0;kk<N-M;kk++)
mix1(?N-?M, List, [], Acc) ->
    {List, Acc};
mix1(Kk, List, ListpM, Acc) ->
    MtKK = hd(List),
    MtKKp1 = hd(tl(List)),
    MtKKpM = hd(ListpM),

    %y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
    Y = (MtKK band ?UPPER_MASK) bor (MtKKp1 band ?LOWER_MASK),

    %mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
    NewMtKK = MtKKpM bxor (Y bsr 1) bxor mag01(Y band 1),

    mix1(Kk + 1, tl(List), tl(ListpM), [NewMtKK | Acc]).

%for (;kk<N-1;kk++)
mix2(?N-1, List, _, []) ->
    List;
mix2(?N-1, List, _, [Elem | Acc]) ->
    mix2(?N-1, [Elem | List], x, Acc);
mix2(Kk, List, ListpMtN, Acc) ->
    MtKKpMtN = hd(ListpMtN),
    MtKK = hd(List),
    MtKKp1 = hd(tl(List)),

    %y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
    Y = (MtKK band ?UPPER_MASK) bor (MtKKp1 band ?LOWER_MASK),

    %mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
    NewMtKK = MtKKpMtN bxor (Y bsr 1) bxor mag01(Y band 1),

    NewListpMtN = case tl(ListpMtN) of
        [] ->
            lists:nthtail(?N-?M, lists:reverse([NewMtKK | Acc]));
        _ ->
            tl(ListpMtN)
    end,
    mix2(Kk + 1, tl(List), NewListpMtN, [NewMtKK | Acc]).



%
% @spec genrand_int32(State::State) -> {Int32, NewState}
%       State = ran_mt()
%       Int32 = integer()
%       NewState = ran_mt()
%
% @equiv genrand(State)
%
genrand_int32(Rand) ->
    genrand(Rand).



%
% @spec genrand_int31(State::State) -> {Int31, NewState}
%       State = ran_mt()
%       Int31 = integer()
%       NewState = ran_mt()
%
% @doc  Generates a random number on [0,0x7fffffff]-interval
%
genrand_int31(Rand) ->
    %return (long)(genrand_int32()>>1);
    {Int32, NewRand} = genrand(Rand),
    {Int32 bsr 1, NewRand}.



%
% @spec genrand_real1(State::State) -> {Real1, NewState}
%       State = ran_mt()
%       Real1 = float()
%       NewState = ran_mt()
%
% @doc  Generates a random number on [0,1]-real-interval
%
genrand_real1(Rand) ->
    % divided by 2^32-1 
    %return genrand_int32()*(1.0/4294967295.0); 
    {Int32, NewRand} = genrand(Rand),
    {Int32*(1.0/4294967295.0), NewRand}.



%
% @spec genrand_real2(State::State) -> {Real1, NewState}
%       State = ran_mt()
%       Real1 = float()
%       NewState = ran_mt()
%
% @doc  Generates a random number on [0,1)-real-interval
%
genrand_real2(Rand) ->
    % divided by 2^32 
    %return genrand_int32()*(1.0/4294967296.0); 
    {Int32, NewRand} = genrand(Rand),
    {Int32*(1.0/4294967296.0), NewRand}.



%
% @spec genrand_real3(State::State) -> {Real1, NewState}
%       State = ran_mt()
%       Real1 = float()
%       NewState = ran_mt()
%
% @doc Generates a random number on (0,1)-real-interval
%
genrand_real3(Rand) ->
    % divided by 2^32
    %return (((double)genrand_int32()) + 0.5)*(1.0/4294967296.0);
    {Int32, NewRand} = genrand(Rand),
    {(Int32 + 0.5)*(1.0/4294967296.0), NewRand}.



%
% @spec genrand_res53(State::State) -> {Real1, NewState}
%       State = ran_mt()
%       Real1 = float()
%       NewState = ran_mt()
%
% @doc  Generates a random number on [0,1) with 53-bit resolution
%
genrand_res53(Rand) ->
    %unsigned long a=genrand_int32()>>5, b=genrand_int32()>>6; 
    %return(a*67108864.0+b)*(1.0/9007199254740992.0);
    {Int32A, Rand1} = genrand(Rand),
    {Int32B, NewRand} = genrand(Rand1),
    {((Int32A bsr 5)*67108864.0+(Int32B bsr 6))*(1.0/9007199254740992.0), NewRand}.



% These real versions are due to Isaku Wada, 2002/01/09 added



% From here on is a compatibility layer to the random module in stdlib.

%
% @spec seed0() -> ran_mt()
%
% @doc  Returns the default state.
%
seed0() ->
    init({3172, 9814, 20125}).



%
% @spec seed() -> ran_mt()
%
% @doc  Seeds random number generation with default (fixed) values in the
%       process dictionary, and returns the old state.
%
seed() ->
    put(random_mt_seed, seed0()).



%
% @spec seed(A1, A2, A3) -> ran_mt()
%
% @doc  Seeds random number generation with integer values in the process
%       dictionary, and returns the old state.
%
seed(A1, A2, A3) ->
    put(random_mt_seed, init([A1, A2, A3])).



%
% @spec uniform() -> float()
%
% @doc  Returns a random float uniformly distributed between 0.0 and 1.0,
%       updating the state in the process dictionary.
%
uniform() ->
    Rand = case get(random_mt_seed) of
        undefined -> seed0();
        Tuple -> Tuple
    end,
    {Num, NextRand} = genrand_real1(Rand),
    put(random_mt_seed, NextRand),
    Num.



%
% @spec uniform(N) -> integer()
%       N = integer()
%
% @doc  Given an integer N >= 1, uniform/1 returns a random integer uniformly
%       distributed between 1 and N, updating the state in the process
%       dictionary.
%
uniform(N) when N >= 1 ->
    trunc(uniform() * N) + 1.



%
% @spec uniform_s(State0) -> {float(), State1}
%       State0 = ran_mt()
%       State1 = ran_mt()
%
% @doc  Given a state, uniform_s/1 returns a random float uniformly distributed
%       between 0.0 and 1.0, and a new state.
%
uniform_s(State0) ->
    genrand_real1(State0).



%
% @spec uniform_s(N, State0) -> {integer, State1}
%       N = integer()
%       State0 = ran_mt()
%       State1 = ran_mt()
%
% @doc  Given an integer N >= 1 and a state, uniform_s/2 returns a random
%       integer uniformly distributed between 1 and N, and a new state.
%
uniform_s(N, Rand) ->
    {Num, NextRand} = genrand_real1(Rand),
    {trunc(Num * N) + 1, NextRand}.



